"""Rest class definition"""

import logging
from time import perf_counter
from typing import Dict, Tuple

import requests
from requests.auth import HTTPBasicAuth
from requests_futures.sessions import FuturesSession

# Getting the logger reference
logger = logging.getLogger()


class Rest:
    """Handeling of REST requests"""

    def __init__(self, username: str = '', api_token: str = '', server_url: str = '', session=None) -> None:
        """TODO Docstring

        Args:
            TODO

        Details:
            Troubleshoot: nettop -p Python

        Returns:
            TODO
        """
        # Request session
        if not session:
            logger.debug('Starting new requests session (Type: FuturesSession) ...')
            self.request_session = FuturesSession(max_workers=16)
        else:
            # Convert to future session
            logger.debug('Converting request session to FutureSession ...')
            self.request_session = FuturesSession(session=session, max_workers=16)

        # Authentication passed
        self.username: str = username
        self.api_token: str = api_token
        self.server_url: str = server_url

        # Flag signaling if this object has authentication credentials to server
        self.has_credentials = False

    def set_credentials(self, username: str, api_token: str, server_url: str) -> None:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        self.username = username
        self.api_token = api_token
        self.server_url = server_url.strip('/') + '/'
        self.has_credentials = True

    def get_server_url(self) -> str:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        return self.server_url

    def get_active_session(self) -> object:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        return self.request_session

    def is_reachable(self, server_url: str = '', timeout: int = 5) -> bool:
        """Check if the server is reachable

        Details:
            `/login` endpoint used to verify

        Args:
            server_url: URL of server to check

        Returns:
            True if reachable, else False
        """
        if not server_url:
            server_url = self.server_url.strip('/') + '/login'

        logger.debug(f'Checking if server is reachable: "{server_url}" ...')

        request_success = self.request(target=server_url,
                                       is_endpoint=False,
                                       request_type='head',
                                       auth_needed=False,
                                       timeout=timeout)[1]

        if request_success:
            logger.debug('Successfully found server is reachable')
            return True
        else:
            logger.debug('Failed. Server cannot be reached or is offline')
            return False

    def request(self,
                target: str,
                request_type: str,
                is_endpoint: bool = True,
                json_content: bool = True,
                auth: Tuple = None,
                auth_needed: bool = True,
                new_session: bool = False,
                params: dict = {},
                data: dict = {},
                json_data: dict = {},
                headers: dict = {},
                timeout: int = 10,
                allow_redirect: bool = True) -> Tuple[Dict, Dict, bool]:
        """Utility method for a single REST requests

        Details: Currently supported GET, POST, HEAD

        **TODO**: Refactor/Rework this method. Too bloated. Take appart into multiple methods!

        Args:
            target         : Request URL target. Does not include server_url
            is_endpoint    : If True, add the object-stored server URL address, else do not
            request_type   : Type of request. Currently `get`, `post`, `head` only
            json_content   : If True, parse as json/dict, else return raw content text
            auth           : Credentials in (username, password) format
            auth_needed    : If True, use credentials, else do not
            new_session    : If True, create a new connection sessions, else re-use previous/default session
            params         : Parameters passed with the request
            data           : Data passed with the request
            json_data      : JSON data passed with the request
            headers        : Headers passed with the request
            timeout        : Number of seconds to wait for request
            allow_redirect : If True, allow request redirection to other URLs

        Returns:
            Tuple of return content, return header, return success
        """
        # Constructing the request URL
        if is_endpoint:
            request_url = self.server_url.strip('/') + '/' + target.strip('/')
        else:
            request_url = target

        logger.debug(f'Request URL: {request_url}')

        # Get credentials if needed
        if auth_needed:
            if not auth:
                auth = HTTPBasicAuth(self.username, self.api_token)

        # Use a connection session if possible
        if not self.request_session or new_session:
            logger.debug('Starting new requests session')
            self.request_session = FuturesSession(max_workers=16)

        # Making the request
        start_time = perf_counter()
        try:
            if request_type.lower() == 'get':
                response = self.request_session.get(request_url,
                                                    params=params,
                                                    data=data,
                                                    json=json_data,
                                                    headers=headers,
                                                    auth=auth,
                                                    timeout=timeout,
                                                    allow_redirects=allow_redirect)
            elif request_type.lower() == 'post':
                response = self.request_session.post(request_url,
                                                     params=params,
                                                     data=data,
                                                     json=json_data,
                                                     headers=headers,
                                                     auth=auth,
                                                     timeout=timeout,
                                                     allow_redirects=allow_redirect)
            elif request_type.lower() == 'head':
                response = self.request_session.head(request_url,
                                                     params=params,
                                                     data=data,
                                                     json=json_data,
                                                     headers=headers,
                                                     auth=auth,
                                                     timeout=timeout,
                                                     allow_redirects=allow_redirect)
            elif request_type.lower() == 'delete':
                response = self.request_session.delete(request_url,
                                                       params=params,
                                                       data=data,
                                                       json=json_data,
                                                       headers=headers,
                                                       auth=auth,
                                                       timeout=timeout,
                                                       allow_redirects=allow_redirect)
            else:
                logger.debug(f'Request type "{request_type}" not recognized')
                return {}, {}, False
        except (requests.exceptions.ConnectionError, requests.exceptions.InvalidSchema,
                requests.exceptions.RequestException) as error:
            logger.debug(f'Failed to make request. Exception: {error}')
            return {}, {}, False

        # Wait on the response to complete and get result
        try:
            if hasattr(response, 'result'):
                response = response.result()
        except (requests.exceptions.RequestException, Exception) as error:
            logger.debug(f'Failed to make request. Exception: {error}')
            return {}, {}, False

        # For Debug Purposes
        # print(html_clean(str(response.content)))

        # Logging any request redirects
        if response.history:
            for i, redirect_hop in enumerate(response.history):
                logger.debug('Request redirects:')
                logger.debug(
                    f'    - {i+1}/{len(response.history)}: {redirect_hop.request.method} - {redirect_hop.url} - Status: {redirect_hop.status_code}'
                )

        redirect_methods = ": " + " --> ".join([r.request.method for r in response.history] +
                                               [response.request.method]) if response.history else ''
        response_content_type = response.headers["Content-Type"] if "Content-Type" in response.headers else "N/A"
        response_content_len = response.headers["Content-Length"] if "Content-Length" in response.headers else "N/A"

        # Final request report
        logger.debug('Request summary:')
        logger.debug(f'   - Request:          {response.request.method} - {request_url}')
        logger.debug(f'   - Redirects:        {"Allow" if allow_redirect else "Block"} {redirect_methods}')
        logger.debug(f'   - Elapsed time:     {perf_counter() - start_time:.3f} seconds')
        logger.debug(f'   - Response headers: {response_content_type} (Content Length Bytes: {response_content_len})')
        logger.debug(f'   - Status code:      {response.status_code} ({response.reason})')

        # Check for permission denied
        if response.status_code in [401, 403, 405]:
            logger.debug('PERMISSION DENIED - Request denied due to insufficient privileges')

        # Check conflict
        if response.status_code in [409]:
            logger.debug('CONFLICT - Request failed due to conflict. Possible duplicate entry')

        # If a head request, only return headers
        if request_type.lower() == 'head':
            return {}, response.headers, response.ok

        # Check the return status code
        if not response.ok:
            logger.debug(
                f'Failed to make {request_type.upper()} request "{request_url}". Server code: {response.status_code}')
            if response.headers:
                logger.debug(f'Response headers: {response.headers}')
            return {}, {}, False

        # Get the return content and format it
        return_content = {}
        if response.content:
            if json_content:
                # Check for json parsing errors
                try:
                    return_content = response.json()
                except Exception as error:
                    # TODO: Specify json parse error
                    logger.debug(f"Failed to parse request return as JSON. Possible HTML content. Exception: {error})")
            else:
                return_content = response.text
        else:
            logger.debug(f'No content received from {request_type.upper()} request: {request_url}')

        return return_content, response.headers, True



"""Node click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import node
from yojenkins.cli import cli_decorators, cli_node
from yojenkins.cli.cli_utility import set_debug_log_level


@node.command(short_help='\tNode information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('-d', '--depth', type=int, default=0, required=False, help='Search depth from root directory')
def info(debug, pretty, yaml, xml, toml, profile, name, depth):
    """Node information"""
    set_debug_log_level(debug)
    cli_node.info(pretty, yaml, xml, toml, profile, name, depth)


@node.command(short_help='\tNode status')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
def status(debug, profile, name):
    """Node status"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')


@node.command(short_help='\tList all nodes')
@cli_decorators.debug
@cli_decorators.profile
@cli_decorators.format_output
@cli_decorators.list
@click.option('-d', '--depth', type=int, default=0, required=False, help='Search depth from root directory')
def list(debug, profile, pretty, yaml, xml, toml, list, depth):
    """List all nodes"""
    set_debug_log_level(debug)
    cli_node.list(pretty, yaml, xml, toml, list, profile, depth)


@node.command(short_help='\tPrepare a remote machine to become a node')
@cli_decorators.debug
def prepare(debug):
    """Prepare a remote machine to become a node"""
    set_debug_log_level(debug)
    cli_node.prepare()


@node.command(short_help='\tSetup a local or remote persistent node')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.argument('host', nargs=1, type=str, required=True)
@click.argument('credential', nargs=1, type=str, required=True)
@click.option('--description', type=str, required=False, help='Node description')
@click.option('--executors',
              default=1,
              type=click.IntRange(1, 100),
              required=False,
              show_default=True,
              help='Number of executors on node')
@click.option('--labels', type=str, required=False, help='Labels applied to agent [default: NAME]')
@click.option('--mode',
              type=click.Choice(['normal', 'exclusive'], case_sensitive=False),
              default='normal',
              show_default=True,
              required=False,
              help='Available to all or to specified jobs')
@click.option('--remote-java-dir',
              default="/usr/bin/java",
              show_default=True,
              type=str,
              required=False,
              help='Location of Java binary')
@click.option('--remote-root-dir',
              default="/home/jenkins",
              show_default=True,
              type=str,
              required=False,
              help='Directory where node work is kept')
@click.option('--retention',
              type=click.Choice(['always', 'demand'], case_sensitive=False),
              default='always',
              show_default=True,
              required=False,
              help='Always on or offline when not in use')
@click.option('--ssh-port',
              default=22,
              show_default=True,
              type=click.IntRange(1, 64738),
              required=False,
              help='SSH port to target')
@click.option('--ssh-verify',
              type=click.Choice(['known', 'trusted', 'none'], case_sensitive=False),
              default='trusted',
              show_default=True,
              required=False,
              help='SSH verification strategy')
# @click.option('--config-file', type=click.Path(file_okay=True, dir_okay=False), required=False, help='Path to local XML file defining agent')
def create_permanent(debug, profile, **kwargs):
    """
    This command sets up a local or remote node on a virtual machine, container,
    or physical machine, connecting with SSH. The target system must have the following:

    \b
    - A working SSH server installed, running, and accessible from main server
    - Java installed

    This command only sets the node up, but it does not monitor to see if the agent
    has successfully connected. You will either need to manually check the node in the Jenkins UI,
    or you can use: "yojenkins node status NAME"

    ARGUMENTS:

    \b
      NAME:        Name of the node
      HOST:        Hostname or IP address of the node
      CREDENTIAL:  SSH type credential in Jenkins

    EXAMPLES:

    \b
    - yojenkins node create-permanent my-node 192.168.0.23 my-cred --description "Yo new node"
    - yojenkins node create-permanent my-node 192.168.0.23 15ad1f93-dc24-4f71-b92b-18ae9b13b1d0
    - yojenkins node create-permanent "Node 1" ey-yo.com my-cred --labels label1,label2,label3
    """
    set_debug_log_level(debug)
    cli_node.create_permanent(profile, **kwargs)


@node.command(short_help='\tSetup a local or remote ephemeral/as-needed node')
@cli_decorators.debug
def create_ephemeral(debug):
    """Setup a local or remote ephemeral/as-needed node"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')


@node.command(short_help='\tDelete a node')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
def delete(debug, profile, name):
    """Delete a node"""
    set_debug_log_level(debug)
    cli_node.delete(profile, name)


@node.command(short_help='\tDisable a node')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('--message', type=str, required=False, help='Message for disabling node')
def disable(debug, profile, name, message):
    """Disable a node"""
    set_debug_log_level(debug)
    cli_node.disable(profile, name, message)


@node.command(short_help='\tEnable a node')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('--message', type=str, required=False, help='Message for enabling node')
def enable(debug, profile, name, message):
    """Enable a node"""
    set_debug_log_level(debug)
    cli_node.enable(profile, name, message)


@node.command(short_help='\tGet node configuration')
@cli_decorators.debug
@cli_decorators.format_output
@click.option('-j',
              '--json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Output config in JSON format')
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('--filepath',
              type=click.Path(file_okay=True, dir_okay=True),
              required=False,
              help='File/Filepath to write configurations to')
def config(debug, pretty, yaml, xml, toml, json, profile, name, filepath):
    """Get node configuration"""
    set_debug_log_level(debug)
    cli_node.config(pretty, yaml, xml, toml, json, profile, name, filepath)


@node.command(short_help='\tReconfigure the node')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.option('--config-file',
              type=click.Path(file_okay=True, dir_okay=True),
              required=True,
              help='Path to local config file defining node')
@click.option('--config-is-json',
              type=bool,
              default=False,
              show_default=True,
              required=False,
              is_flag=True,
              help='The specified file is in JSON format')
def reconfig(debug, profile, name, config_file, config_is_json):
    """Reconfigure the node"""
    set_debug_log_level(debug)
    cli_node.reconfig(profile, name, config_file, config_is_json)


@node.command(short_help='\tNode logs')
@cli_decorators.debug
@cli_decorators.profile
def logs(debug, profile):
    """Node logs"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')



"""Credential click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import credential
from yojenkins.cli import cli_credential, cli_decorators
from yojenkins.cli.cli_utility import set_debug_log_level


@credential.command(short_help='\tList credentials')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.list
@cli_decorators.profile
# @click.argument('folder', nargs=1, type=str, default="root", required=False)
@click.option('--folder', type=str, default="root", show_default=True, required=False, help='Credentials folder')
@click.option('--domain', type=str, default="global", show_default=True, required=False, help='Credentials domain')
@click.option('--keys',
              type=str,
              default="all",
              show_default=True,
              required=False,
              help='Credential info keys to return [ie. displayName,id,etc]')
def list(debug, pretty, yaml, xml, toml, list, profile, folder, domain, keys):
    """List credentials"""
    set_debug_log_level(debug)
    cli_credential.list(pretty, yaml, xml, toml, list, profile, folder, domain, keys)


@credential.command(short_help='\tCredential information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('credential', nargs=1, type=str, required=True)
@click.option('--folder', type=str, default="root", show_default=True, required=False, help='Credential folder')
@click.option('--domain', type=str, default="global", show_default=True, required=False, help='Credential domain')
def info(debug, pretty, yaml, xml, toml, profile, credential, folder, domain):
    """Credential information"""
    set_debug_log_level(debug)
    cli_credential.info(pretty, yaml, xml, toml, profile, credential, folder, domain)


@credential.command(short_help='\tGet credential configuration')
@cli_decorators.debug
@cli_decorators.format_output
@click.option('-j',
              '--json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Output config in JSON format')
@cli_decorators.profile
@click.argument('credential', nargs=1, type=str, required=True)
@click.option('--folder', type=str, default="root", show_default=True, required=False, help='Credential folder')
@click.option('--domain', type=str, default="global", show_default=True, required=False, help='Credential domain')
@click.option('--filepath',
              type=click.Path(file_okay=True, dir_okay=True),
              required=False,
              help='File/Filepath to write configurations to')
def config(debug, pretty, yaml, xml, toml, json, profile, credential, folder, domain, filepath):
    """Get credential configuration"""
    set_debug_log_level(debug)
    cli_credential.config(pretty, yaml, xml, toml, json, profile, credential, folder, domain, filepath)


@credential.command(short_help='\tCredential type template to create a credential')
@cli_decorators.debug
@cli_decorators.format_output
@click.option('-j',
              '--json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Output config in JSON format')
@cli_decorators.profile
@click.argument('type',
                type=click.Choice(['user-pass', 'ssh-key', 'secret-text'], case_sensitive=False),
                default='user-pass',
                required=True)
@click.option('--filepath',
              type=click.Path(file_okay=True, dir_okay=True),
              required=False,
              help='File/Filepath to write template to')
def get_template(debug, pretty, yaml, xml, toml, json, profile, type, filepath):
    """Credential type template to create a credential"""
    set_debug_log_level(debug)
    cli_credential.get_template(pretty, yaml, xml, toml, json, profile, type, filepath)


@credential.command(short_help='\tCreate new credentials')
@cli_decorators.debug
@cli_decorators.profile
# @click.argument('name', nargs=1, type=str, required=True)
@click.argument('config-file', nargs=1, type=click.Path(exists=True), required=True)
@click.option('--folder', type=str, default="root", show_default=True, required=False, help='Credential folder')
@click.option('--domain', type=str, default="global", show_default=True, required=False, help='Credential domain')
def create(debug, profile, config_file, folder, domain):
    """Create new credentials"""
    set_debug_log_level(debug)
    cli_credential.create(profile, config_file, folder, domain)


@credential.command(short_help='\tRemove credentials')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('credential', nargs=1, type=str, required=True)
@click.option('--folder', type=str, default="root", show_default=True, required=False, help='Credential folder')
@click.option('--domain', type=str, default="global", show_default=True, required=False, help='Credential domain')
def delete(debug, profile, credential, folder, domain):
    """Remove credentials"""
    set_debug_log_level(debug)
    cli_credential.delete(profile, credential, folder, domain)


@credential.command(short_help='\tReconfigure existing credentials')
@cli_decorators.debug
def update(debug):
    """Reconfigure existing credentials"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')


@credential.command(short_help='\tMove a credential to another folder/domain')
@cli_decorators.debug
def move(debug):
    """Move a credential to another folder/domain"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')



"""Server class definition"""

import logging
from typing import Dict, List, Tuple

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out

# Getting the logger reference
logger = logging.getLogger()


class Server():
    """TODO Server"""

    def __init__(self, rest: object, auth: object) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest
        self.auth = auth

        self.server_base_url = auth.jenkins_profile['jenkins_server_url']

    def info(self) -> Dict:
        """Get the server information

        Details: Targeting the server that is specified in the selected profile

        Args:
            None

        Returns:
            Server information
        """
        server_info, _, success = self.rest.request('api/json', 'get')
        if not success:
            fail_out('Failed to fetch server information')

        return server_info

    def people(self) -> Tuple[list, list]:
        """Get the list of people/accounts on the server

        Args:
            None

        Returns:
            List of users/accounts, list of usernames
        """
        logger.debug(f'Getting all people/users for "{self.server_base_url}" ...')

        people_info, _, success = self.rest.request('asynchPeople/api/json?depth=1', 'get', is_endpoint=True)
        if not success:
            fail_out('Failed to fetch server people/users information')

        try:
            people_info = people_info['users']
            people_info_list = [f"{p['user']['fullName']}" for p in people_info]
        except KeyError as error:
            fail_out(f'Failed to parse server people/users information. Specific keys not found: {error}')

        return people_info, people_info_list

    def queue_info(self) -> Dict:
        """Get all the jobs stuck in the server queue

        (Potentially move to jobs or build section)

        Args:
            None

        Returns:
            Server queue information
        """
        # TODO: Combine with server_queue_list adding a list argument

        logger.debug(f'Requesting build queue info for "{self.server_base_url}" ...')

        # Making the request
        server_queue_info, _, success = self.rest.request('queue/api/json', 'get')
        if not success:
            fail_out('Failed to get server queue info')

        return server_queue_info

    def queue_list(self) -> List[str]:
        """Get all list of all the jobs stuck in the server queue

        (Potentially move to jobs or build section)

        Args:
            None

        Returns:
            List of urls of jobs stuck in server queue
        """
        server_queue_info = self.queue_info()

        queue_list = []
        queue_job_url_list = []
        for queue_item in server_queue_info['items']:
            queue_list.append(queue_item)
            if 'url' in queue_item['task']:
                queue_job_url_list.append(queue_item['task']['url'])

        return queue_list

    def plugin_list(self) -> Tuple[list, list]:
        """Get the list of plugins installed on the server

        Args:
            None

        Returns:
            List of plugins, information list and URL list
        """
        logger.debug(f'Getting all installed server plugins for "{self.server_base_url}" ...')

        plugins_info, _, success = self.rest.request('pluginManager/api/json?depth=2', 'get', is_endpoint=True)
        if not success:
            fail_out('Failed to fetch server plugin information')

        try:
            plugins_info = plugins_info['plugins']
            plugin_info_list = [f"{p['longName']} - {p['shortName']} - {p['version']}" for p in plugins_info]
        except KeyError as error:
            fail_out(f'Failed to parse server plugin information. Specific keys not found: {error}')

        return plugins_info, plugin_info_list

    def browser_open(self) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Opening in server home page in browser: "{self.server_base_url}" ...')
        success = utility.browser_open(url=self.server_base_url)
        if not success:
            fail_out('Failed to open server home page in browser')
        logger.debug('Successfully opened in web browser')

        return success

    def restart(self, force: bool = True) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug("Initiating server force restart ..." if force else "Initiating server safe restart ...")
        success = self.rest.request('restart' if force else 'safeRestart',
                                    'post',
                                    is_endpoint=True,
                                    json_content=True,
                                    allow_redirect=False)[2]
        if not success:
            fail_out('Failed to initiate server restart')

        return success

    def shutdown(self, force: bool = True) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug("Initiating server force shutdown ..." if force else "Initiating server safe shutdown ...")
        success = self.rest.request('exit' if force else 'safeExit',
                                    'post',
                                    is_endpoint=True,
                                    json_content=False,
                                    allow_redirect=False)[2]
        if not success:
            fail_out('Failed to initiate server shutdown')

        return success

    def quiet(self, off: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug("Disabeling server quiet mode ..." if off else "Enabling server quiet mode ...")
        success = self.rest.request('cancelQuietDown' if off else 'quietDown',
                                    'post',
                                    is_endpoint=True,
                                    json_content=True,
                                    allow_redirect=False)[2]
        if not success:
            fail_out(f'Failed to {"disable" if off else "enable"} server quiet mode')

        return success



"""Status class definition"""

from enum import Enum


class Status(Enum):
    """Enum of Jenkins statuses as described by Jenkins return

    Usage Examples:
        - `BuildStatus.RUNNING.value`
        - `if some_text in Status.RUNNING.value:`
    """
    RUNNING = ['RUNNING', 'IN_PROGRESS']
    SUCCESS = ['SUCCESS', 'SUCCEEDED']
    FAILURE = ['FAILURE', 'FAILED', 'FAIL']
    QUEUED = ['QUEUED']
    ABORTED = ['ABORTED']
    UNSTABLE = ['UNSTABLE']
    PAUSED_INPUT = ['PAUSED_PENDING_INPUT']
    NOT_FOUND = ['NOT FOUND OR STARTED']
    NOT_RUN = ['NOT_EXECUTED', 'NOT_RUN']
    NONE = [None]
    UNKNOWN = ['UNKNOWN']


class BuildStatus(Enum):
    """Enum of Jenkins status for build

    Details: This enum references Status enum above and
             picks the status text that belongs to it

    Usage Examples:  `BuildStatus.RUNNING.value`
    """
    RUNNING = Status.RUNNING.value[0]
    SUCCESS = Status.SUCCESS.value[0]
    FAILURE = Status.FAILURE.value[0]
    QUEUED = Status.QUEUED.value[0]
    ABORTED = Status.ABORTED.value[0]
    UNSTABLE = Status.UNSTABLE.value[0]
    PAUSED_INPUT = Status.PAUSED_INPUT.value[0]
    NOT_FOUND = Status.NOT_FOUND.value[0]
    NOT_RUN = Status.NOT_RUN.value[0]
    NONE = Status.NONE.value[0]
    UNKNOWN = Status.UNKNOWN.value[0]


class StageStatus(Enum):
    """Enum of Jenkins status for stage

    Details: This enum references Status enum above and
             picks the status text that belongs to it

    Usage Examples:  `StageStatus.RUNNING.value`
    """
    RUNNING = Status.RUNNING.value[1]  # <-- Note 1
    SUCCESS = Status.SUCCESS.value[0]
    FAILURE = Status.FAILURE.value[0]
    QUEUED = Status.QUEUED.value[0]
    ABORTED = Status.ABORTED.value[0]
    UNSTABLE = Status.UNSTABLE.value[0]
    PAUSED_INPUT = Status.PAUSED_INPUT.value[0]
    NOT_FOUND = Status.NOT_FOUND.value[0]
    NOT_RUN = Status.NOT_RUN.value[0]
    NONE = Status.NONE.value[0]
    UNKNOWN = Status.UNKNOWN.value[0]


class Color(Enum):
    """Enum of Jenkins status color name

    Usage Examples:  `Color.RUNNING.value`
    """
    ITEMS = {
        'RUNNING': 'normal',
        'SUCCESS': 'green',
        'FAILURE': 'red',
        'QUEUED': 'normal',
        'ABORTED': 'magenta',
        'UNSTABLE': 'orange',
        'PAUSED_INPUT': 'cyan',
        'NOT_FOUND': 'normal',
        'NOT_RUN': 'grey-dark',
        'NONE': 'normal',
        'UNKNOWN': 'normal'
    }


class Sound(Enum):
    """Enum of local sound filename matching Jenkins status

    Details:
        - Sound source: https://www.zapsplat.com/

    Usage Examples:  `Sound.RUNNING.value`
    """
    ITEMS = {
        'RUNNING': '',
        'SUCCESS': 'positive_alert_notification_musical_short_marimba_process_finished.wav',
        'FAILURE': 'negative_ui_ping_chime_mallet_like_error.wav',
        'QUEUED': '',
        'ABORTED': 'negative_ui_ping_chime_mallet_like.wav',
        'UNSTABLE': 'negative_notification_mallet_wooden_short_alert.wav',
        'PAUSED_INPUT': 'positive_alert_mallet_wood_simple_musical_003.wav',
        'NOT_FOUND': '',
        'NOT_RUN': '',
        'NONE': '',
        'UNKNOWN': ''
    }



"""Node class definition"""

import json
import logging
import os
from typing import Tuple

import xmltodict

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out, print2
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses

# Getting the logger reference
logger = logging.getLogger()


class Node():
    """TODO Node"""

    def __init__(self, rest: object) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest

    def info(self, node_name: str, depth: int = 0) -> Tuple[list, list]:
        """TODO Docstring

        Details: TODO

        Args:
            node_name: TODO
            depth: TODO

        Returns:
            TODO
        """
        logger.debug(f'Getting info for node: {node_name} ...')
        node_name = "(master)" if node_name == 'master' else node_name  # Special case
        node_info, _, success = self.rest.request(target=f"computer/{node_name}/api/json?depth={depth}",
                                                  request_type='get',
                                                  is_endpoint=True,
                                                  json_content=True)
        if not success:
            fail_out(f'Failed to find node info for "{node_name}"')

        return node_info

    def list(self, depth: int = 0) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug('Getting a list of all nodes ...')
        nodes_info, _, success = self.rest.request(target=f"computer/api/json?depth={depth}",
                                                   request_type='get',
                                                   is_endpoint=True,
                                                   json_content=True)
        if not success:
            fail_out('Failed to get any nodes')

        if "computer" not in nodes_info:
            fail_out('Failed to find "computer" section in return content')

        node_list, node_list_name = utility.item_subitem_list(
            item_info=nodes_info,
            get_key_info='displayName',
            item_type=JenkinsItemClasses.NODE.value['item_type'],
            item_class_list=JenkinsItemClasses.NODE.value['class_type'])

        logger.debug(f'Number of nodes found: {len(node_list)}')
        logger.debug(f'Node names: {node_list_name}')

        return node_list, node_list_name

    def create_permanent(self, **kwargs) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Check if the node already exists. Call exists() method. Need info() method.
        # TODO: Check if credentials exists on server. Check type being SSH

        # Checking name for special characters
        if utility.has_special_char(kwargs['name']):
            fail_out('Provided node name contains special characters')

        # Processing labels
        if kwargs['labels']:
            labels = utility.parse_and_check_input_string_list(kwargs['labels'], " ")
            if not labels:
                return False
        else:
            # If not list passed, use the node name as label
            labels = kwargs['name']

        logger.debug('Creating and configuring a new permanent node/agent ...')
        logger.debug(f'    - Name:       {kwargs["name"]}')
        logger.debug(f'    - Host:       {kwargs["host"]}')
        logger.debug('    - Connection: SSH')

        # SSH Connection verification strategy
        if kwargs['ssh_verify'] == 'known':
            ssh_verify = {"stapler-class": "hudson.plugins.sshslaves.verifiers.KnownHostsFileKeyVerificationStrategy"}
        elif kwargs['ssh_verify'] == 'trusted':
            ssh_verify = {"stapler-class": "hudson.plugins.sshslaves.verifiers.ManuallyTrustedKeyVerificationStrategy"}
        elif kwargs['ssh_verify'] == 'provided':
            ssh_verify = {
                "stapler-class": "hudson.plugins.sshslaves.verifiers.ManuallyProvidedKeyVerificationStrategy"
            }
        elif kwargs['ssh_verify'] == 'none':
            ssh_verify = {"stapler-class": "hudson.plugins.sshslaves.verifiers.NonVerifyingKeyVerificationStrategy"}

        ssh_launcher = {
            "stapler-class": "hudson.plugins.sshslaves.SSHLauncher",
            "host": kwargs['host'],
            "includeUser": False,
            "credentialsId": kwargs['credential'],
            "sshHostKeyVerificationStrategy": ssh_verify,
            "port": kwargs['ssh_port'],
            "javaPath": kwargs['remote_java_dir'],
            "jvmOptions": "",
            "prefixStartSlaveCmd": "",
            "suffixStartSlaveCmd": "",
            "launchTimeoutSeconds": 60,
            "maxNumRetries": 5,
            "retryWaitTime": 15,
            "tcpNoDelay": True,
            "workDir": ""
        }

        json_params = {
            'nodeDescription': kwargs['description'],
            'numExecutors': kwargs['executors'],
            'remoteFS': kwargs['remote_root_dir'],
            'labelString': labels,
            'mode': kwargs['mode'].upper(),
            'retentionStrategy': {
                'stapler-class': f'hudson.slaves.RetentionStrategy${kwargs["retention"].capitalize()}'
            },
            'nodeProperties': {
                'stapler-class-bag': True
            },
            'launcher': ssh_launcher
        }

        params = {
            'name': kwargs['name'],
            'type': "hudson.slaves.DumbSlave$DescriptorImpl",
            'json': json.dumps(json_params)
        }

        # Send the request to the server
        success = self.rest.request(target="computer/doCreateItem", request_type='post', is_endpoint=True,
                                    data=params)[2]
        if not success:
            fail_out('Failed to create permanent node')
        logger.debug('Successfully created permanent node')

        return success

    def delete(self, node_name: str) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug(f'Deleting node: {node_name}')
        success = self.rest.request(target=f"computer/{node_name}/doDelete",
                                    request_type='post',
                                    is_endpoint=True,
                                    json_content=False)[2]
        if not success:
            fail_out(f'Failed to delete node "{node_name}"')

        return success

    def disable(self, node_name: str, message: str = None) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug(f'Disabling node: {node_name}')
        logger.debug(f'Message for disabling node: "{message}"')

        # Check if node is disabled already
        node_info = self.info(node_name=node_name)
        if node_info['offline']:
            print2('Node is already disabled')
            return True

        success = self.rest.request(target=f"computer/{node_name}/toggleOffline?offlineMessage={message}",
                                    request_type='post',
                                    is_endpoint=True,
                                    json_content=False)[2]
        if not success:
            fail_out(f'Failed to disable node "{node_name}"')

        return success

    def enable(self, node_name: str, message: str = None) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug(f'Enabling node: {node_name}')
        logger.debug(f'Message for enabling node: "{message}"')

        # Check if node is disabled already
        node_info = self.info(node_name=node_name)
        if not node_info['offline']:
            print2('Node is already enabled')
            return True

        success = self.rest.request(target=f"computer/{node_name}/toggleOffline?offlineMessage={message}",
                                    request_type='post',
                                    is_endpoint=True,
                                    json_content=False)[2]
        if not success:
            fail_out(f'Failed to enable node "{node_name}"')

        return success

    def config(self,
               filepath: str = '',
               node_name: str = '',
               folder_url: str = '',
               opt_json: bool = False,
               opt_yaml: bool = False,
               opt_toml: bool = False) -> Tuple[str, bool]:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug(f'Fetching XML configurations for node: {node_name} ...')
        node_name = "(master)" if node_name == 'master' else node_name  # Special case
        return_content, _, success = self.rest.request(f'computer/{node_name}/config.xml',
                                                       'get',
                                                       json_content=False,
                                                       is_endpoint=True)
        if not success:
            fail_out(f'Failed to fetch node configurations for node "{node_name}"')

        if filepath:
            write_success = utility.write_xml_to_file(return_content, filepath, opt_json, opt_yaml, opt_toml)
            if not write_success:
                fail_out('Failed to write node configurations to file')

        return return_content

    def reconfig(self, node_name: str, config_file: str = None, config_is_json: bool = False) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            name: TODO

        Returns:
            TODO
        """
        logger.debug(f'Reconfiguring node: {node_name} ...')
        logger.debug(f'Using the following specified configuration file: {config_file}')

        logger.debug(f'Checking if file exists: {config_file} ...')
        if not os.path.isfile(config_file):
            fail_out('Specified node configuration file does not exist')

        logger.debug(f'Reading configuration file: {config_file} ...')
        try:
            with open(config_file, 'rb') as file:
                node_config = file.read()
            logger.debug('Successfully read configuration file')
        except (OSError, IOError, PermissionError) as error:
            fail_out(f'Failed to open and read node configuration file. Exception: {error}')

        if config_is_json:
            logger.debug('Converting the specified JSON file to XML format ...')
            try:
                node_config = xmltodict.unparse(json.loads(node_config))
            except Exception as error:
                fail_out(f'Failed to convert the specified JSON file to XML format. Exception: {error}')

        success = self.rest.request(target=f"computer/{node_name}/config.xml",
                                    request_type='post',
                                    is_endpoint=True,
                                    data=node_config.encode('utf-8'),
                                    json_content=False)[2]
        if not success:
            fail_out(f'Failed to reconfigure node "{node_name}"')

        return success



"""Folder class definition"""

import json
import logging
import re
from time import perf_counter
from typing import Dict, Tuple

import xmltodict

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses
from yojenkins.yo_jenkins.jenkins_item_config import JenkinsItemConfig

# Getting the logger reference
logger = logging.getLogger()


class Folder():
    """TODO Folder"""

    def __init__(self, rest, JenkinsSDK) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest
        self.jenkins_sdk = JenkinsSDK

        # Recursive search results
        self.search_results = []
        self.search_items_count = 0

    def __recursive_search(self, search_pattern: str, search_list: list, level: int, fullname: bool = True) -> None:
        """Recursive search method for folders

        Details: Matched pattern findings are storred in the object: `self.search_results`

        Args:
            search_pattern : REGEX pattern to match for each item
            search_list    : List of items
            level          : Current recursion level
            fullname       : Search the entire path of the item, not just the item name

        Returns:
            None
        """
        # Current directory level
        level += 1

        # Loop through all sub-folders
        for list_item in search_list:
            # print(level, "   " * level, list_item['name'])

            # Check if folder
            if list_item['_class'] in JenkinsItemClasses.FOLDER.value['class_type']:

                # Get fullname if specified and available, else get name
                if fullname:
                    dict_key = "fullname" if "fullname" in list_item else "name"
                else:
                    dict_key = 'name'

                # Match the regex pattern
                try:
                    if re.search(search_pattern, list_item[dict_key], re.IGNORECASE):
                        self.search_results.append(list_item)
                except re.error as error:
                    logger.debug(
                        f'Error while applying REGEX pattern "{search_pattern}" to "{list_item[dict_key]}". Exception: {error}'
                    )
                    break

            # Count items searched for record
            self.search_items_count += 1

            # Check if the item has subitems
            if not 'jobs' in list_item:
                continue

            # Keep searching all sub-items for this item. Call itself for some recursion fun
            self.__recursive_search(search_pattern, list_item['jobs'], level, fullname)

    def search(self,
               search_pattern: str,
               folder_name: str = '',
               folder_url: str = '',
               folder_depth: int = 4,
               fullname: bool = True) -> Tuple[list, list]:
        """Search the server for folders matching REGEX pattern

        Args:
            search_pattern : REGEX search pattern to match
            folder_name    : (Optional) Only look within this folder for matching sub-folder using item name
            folder_url     : (Optional) Only look within this folder for matching sub-folder using item URL
            folder_depth   : Number of levels to look through
            fullname       : Search the entire path of the item, not just the item name

        Returns:
            List of folder found. Both, list of info and list of folder URLs
        """
        # Start a timer to time the search
        start_time = perf_counter()
        logger.debug(f'Folder search pattern: {search_pattern}')

        # Get all the jobs
        if folder_name or folder_url:
            # Only recursively search the specified folder name
            logger.debug(f'Searching folder in sub-folder "{folder_name if folder_name else folder_url}"')
            logger.debug('Folder depth does not apply. Only looking in this specific folder for subfolders')
            items = self.item_list(folder_name=folder_name, folder_url=folder_url)[0]
        else:
            # Search entire Jenkins
            logger.debug(f'Searching folder in ALL Jenkins. Folder depth: "{folder_depth}"')
            items = self.jenkins_sdk.get_all_jobs(folder_depth=folder_depth, folder_depth_per_request=20)

        # Search for any matching folders ("jobs")
        self.search_items_count = 0
        self.search_results = []
        self.__recursive_search(search_pattern=search_pattern, search_list=items, level=0, fullname=fullname)

        # Remove duplicates from list (THANKS GeeksForGeeks.org)
        logger.debug('Removing duplicates if needed ...')
        self.search_results = [i for n, i in enumerate(self.search_results) if i not in self.search_results[n + 1:]]

        # Collect URLs only
        folder_search_results_list = []
        for search_result in self.search_results:
            folder_search_results_list.append(search_result['url'])

        # Output search stats
        logger.debug(
            f'Searched folders: {self.search_items_count}. Search time: {perf_counter() - start_time:.3f} seconds')

        return self.search_results, folder_search_results_list

    def info(self, folder_name: str = '', folder_url: str = '') -> Dict:
        """Get the folder information

        Args:
            folder_name : Folder name to get folder information of
            folder_url  : Folder URL to get the folder information of

        Returns:
            Folder information
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_name and not folder_url:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        folder_info, _, success = self.rest.request(folder_url.strip('/') + '/api/json',
                                                    request_type='get',
                                                    is_endpoint=False)
        if not success:
            fail_out(f'Failed to find folder info: {folder_url}')

        # Check if found item type/class
        if folder_info['_class'] not in JenkinsItemClasses.FOLDER.value[
                'class_type'] and JenkinsItemClasses.FOLDER.value['item_type'] not in folder_info:
            fail_out(f'Folder found, but failed to match type/class. This item is "{folder_info["_class"]}"')

        return folder_info

    def subfolder_list(self, folder_name: str = '', folder_url: str = '') -> Tuple[list, list]:
        """Get the list of all sub-folders within the specified folder

        Args:
            folder_name : Folder name to get all its sub-folders
            folder_url  : Folder URL to get all its sub-folders

        Returns:
            List of subfolders, information list and URL list
        """
        logger.debug(f'Getting subfolders for folder name "{folder_name if folder_name else folder_url}" ...')

        # Get the folder information
        folder_info = self.info(folder_name=folder_name, folder_url=folder_url)

        # Extract lists
        sub_folder_list, sub_folder_list_url = utility.item_subitem_list(
            item_info=folder_info,
            get_key_info='url',
            item_type=JenkinsItemClasses.FOLDER.value['item_type'],
            item_class_list=JenkinsItemClasses.FOLDER.value['class_type'])

        logger.debug(f'Number of subfolders found in folder: {len(sub_folder_list)}')
        logger.debug(f'Sub-folders: {sub_folder_list_url}')

        return sub_folder_list, sub_folder_list_url

    def jobs_list(self, folder_name: str = '', folder_url: str = '') -> Tuple[list, list]:
        """Get the list of all jobs within the specified folder

        Args:
            folder_name : Folder name to get all its jobs
            folder_url  : Folder URL to get all its jobs

        Returns:
            List of jobs, information list and URL list
        """
        logger.debug(f'Getting jobs for folder name "{folder_name if folder_name else folder_url}" ...')

        # Get the folder information
        folder_info = self.info(folder_name=folder_name, folder_url=folder_url)

        # Extract lists
        job_list, job_list_url = utility.item_subitem_list(item_info=folder_info,
                                                           get_key_info='url',
                                                           item_type=JenkinsItemClasses.JOB.value['item_type'],
                                                           item_class_list=JenkinsItemClasses.JOB.value['class_type'])

        logger.debug(f'Number of jobs found within folder: {len(job_list)}')
        logger.debug(f'Jobs: {job_list_url}')

        return job_list, job_list_url

    def view_list(self, folder_name: str = '', folder_url: str = '') -> Tuple[list, list]:
        """Get the list of all views within the specified folder

        Args:
            folder_name : Folder name to get all its views
            folder_url  : Folder URL to get all its views

        Returns:
            List of views, information list and URL list
        """
        # Get the folder information
        folder_info = self.info(folder_name=folder_name, folder_url=folder_url)

        logger.debug(f'Getting all view items for folder "{folder_name if folder_name else folder_url}" ...')
        view_list, view_list_url = utility.item_subitem_list(
            item_info=folder_info,
            get_key_info='url',
            item_type=JenkinsItemClasses.VIEW.value['item_type'],
            item_class_list=JenkinsItemClasses.VIEW.value['class_type'])

        logger.debug(f'Number of views found in folder: {len(view_list)}')
        logger.debug(f'Views: {view_list_url}')

        return view_list, view_list_url

    def item_list(self, folder_name: str = '', folder_url: str = '') -> Tuple[list, list]:
        """Get the list of all items within the specified folder

        Args:
            folder_name : Folder name to get all its items
            folder_url  : Folder URL to get all its items

        Returns:
            List of items, information list and URL list
        """
        # TODO: Potentially combine folder_folders_list and folder_jobs_list, folder_view_list with this method

        logger.debug(f'Getting items for folder "{folder_name if folder_name else folder_url}" ...')

        # Get the folder information
        folder_info = self.info(folder_name=folder_name, folder_url=folder_url)

        # Getting all possible Jenkins items listed enum
        all_subitems = [subitem.value for subitem in JenkinsItemClasses]

        # Searching folder info for matches
        all_item_list = []
        all_item_url = []
        for item in all_subitems:
            logger.debug(f'Searching folder for "{item["item_type"]}" items ...')
            item_list, item_list_url = utility.item_subitem_list(item_info=folder_info,
                                                                 get_key_info='url',
                                                                 item_type=item['item_type'],
                                                                 item_class_list=item['class_type'])
            if item_list:
                logger.debug(f'Successfully found {len(item_list)} "{item["item_type"]}" items')
                all_item_list.extend(item_list)
                all_item_url.extend(item_list_url)
            else:
                logger.debug(f'No "{item["item_type"]}" items found in this folder')

        logger.debug(f'Number of folder items found in folder: {len(all_item_list)}')
        logger.debug(f'Items: {all_item_url}')

        return all_item_list, all_item_url

    def browser_open(self, folder_name: str = '', folder_url: str = '') -> bool:
        """Get the list of all items within the specified folder

        Args:
            folder_name : Folder name to get all its items
            folder_url  : Folder URL to get all its items

        Returns:
            True if successfull, else False
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        logger.debug(f'Opening folder in web browser: "{folder_url}" ...')
        success = utility.browser_open(url=folder_url)
        if not success:
            fail_out('Failed to open folder in web browser')
        logger.debug('Successfully opened folder in web browser')

        return success

    def config(self,
               filepath: str = '',
               folder_name: str = '',
               folder_url: str = '',
               opt_json: bool = False,
               opt_yaml: bool = False,
               opt_toml: bool = False) -> str:
        """Get the folder configuration (ie .config.xml)

        Args:
            filepath    : If defined, store fetched data in this file
            folder_name : Folder name to get configurations
            folder_url  : Folder URL to get configurations
            opt_json    : If True, return data as JSON
            opt_yaml    : If True, return data as YAML
            opt_toml    : If True, return data as TOML

        Returns:
            Folder config.xml contents
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        logger.debug(f'Fetching XML configurations for folder: "{folder_url}" ...')
        return_content, _, success = self.rest.request(f'{folder_url.strip("/")}/config.xml',
                                                       'get',
                                                       json_content=False,
                                                       is_endpoint=False)
        if not success:
            fail_out('Failed to get folder configuration')
        logger.debug('Successfully fetched folder configuration')

        if filepath:
            write_success = utility.write_xml_to_file(return_content, filepath, opt_json, opt_yaml, opt_toml)
            if not write_success:
                fail_out('Failed to write configuration file')

        return return_content

    def create(self,
               name: str,
               type: str = 'folder',
               folder_name: str = '',
               folder_url: str = '',
               config: str = 'config.xml',
               config_is_json: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        if not name:
            fail_out('The item name is a blank')
        if utility.has_special_char(name):
            fail_out('The item name contains special characters')

        supported_create_items = ['folder', 'view', 'job']
        if type.strip().lower() not in supported_create_items:
            fail_out(f'Failed to match supported "{type}" items to create: {", ".join(supported_create_items)}')

        if config:
            # Use item configuration from file if provided
            logger.debug(f'Opening and reading "{type}" item configuration file: {config} ...')
            try:
                open_file = open(config, 'rb')
                item_config = open_file.read()
            except (OSError, IOError, PermissionError) as error:
                fail_out(f'Failed to open and read "{type}" item configuration file. Exception: {error}')

            if config_is_json:
                logger.debug('Converting the specified JSON file to XML format ...')
                try:
                    item_config = xmltodict.unparse(json.loads(item_config))
                except ValueError as error:
                    fail_out(f'Failed to convert the specified JSON file to XML format. Exception: {error}')
        else:
            # Use blank item config
            # FIXME: These are not valid XML configs, try json instead
            # FIXME: This does not account for the name of the item
            if type == 'folder':
                endpoint = f'createItem?name={name}'
                item_config = JenkinsItemConfig.FOLDER.value['blank']
                # prefix = JenkinsItemClasses.FOLDER.value['prefix']
            elif type == 'view':
                endpoint = f'createView?name={name}'
                item_config = JenkinsItemConfig.VIEW.value['blank']
                # prefix = JenkinsItemClasses.VIEW.value['prefix']
            elif type == 'job':
                endpoint = f'createItem?name={name}'
                item_config = JenkinsItemConfig.JOB.value['blank']
                # prefix = JenkinsItemClasses.JOB.value['prefix']

        # Checking if the item exists
        if utility.item_exists_in_folder(name, folder_url, type, self.rest):
            fail_out(f'The new "{type}" item "{name}" exists within the folder')

        # Creating the item
        logger.debug(f'Creating "{type}" item "{name}" ...')
        headers = {'Content-Type': 'application/xml; charset=utf-8'}
        success = self.rest.request(f'{folder_url.strip("/")}/{endpoint}',
                                    'post',
                                    data=item_config.encode('utf-8'),
                                    headers=headers,
                                    is_endpoint=False)[2]
        if not success:
            fail_out(f'Failed to create "{type}" item "{name}"')
        logger.debug(f'Successfully created "{type}" item "{name}"')

        # Close the potentially open item configuration file
        try:
            if 'open_file' in locals():
                open_file.close()
        except (OSError, IOError):
            pass

        return success

    def copy(self, original_name: str, new_name: str, folder_name: str = '', folder_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        if not original_name:
            fail_out('The original folder name is a blank')
        if utility.has_special_char(original_name):
            fail_out('The original folder name contains special characters')

        if not new_name:
            fail_out('New folder name is a blank')
        if utility.has_special_char(new_name):
            fail_out('The new folder name contains special characters')

        if not utility.item_exists_in_folder(original_name, folder_url, "folder", self.rest):
            fail_out(f'The original folder "{original_name}" does not exist')

        logger.debug(f'Copying original item "{original_name}" to new item "{new_name}" ...')
        success = self.rest.request(
            f'{folder_url.strip("/")}/createItem?name={new_name}&mode=copy&from={original_name}',
            'post',
            is_endpoint=False)[2]
        if not success:
            fail_out('Failed to copy folder')
        logger.debug('Successfully copied folder')

        return success

    def delete(self, folder_name: str = '', folder_url: str = '') -> bool:
        """Delete folder

        Args:
            folder_name : Folder name to get all its items
            folder_url  : Folder URL to get all its items

        Returns:
            True if successfull, else False
        """
        if not folder_name and not folder_url:
            fail_out('Failed to get folder information. No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        logger.debug(f'Deleting folder: "{folder_url}" ...')
        success = self.rest.request(f'{folder_url.strip("/")}/doDelete', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to delete folder')
        logger.debug('Successfully deleted folder')

        return success



"""Auth click sub-command"""
# pylint: skip-file

import click
from click_help_colors import HelpColorsCommand

from yojenkins.__main__ import auth
from yojenkins.cli import cli_auth, cli_decorators
from yojenkins.cli.cli_utility import set_debug_log_level


@auth.command(short_help='\tConfigure authentication',
              cls=HelpColorsCommand,
              help_options_custom_colors={'--token': 'black'})
@cli_decorators.debug
@click.option('--auth-file',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              is_flag=False,
              help='JSON file containing one or more authentication info')
def configure(debug, auth_file):
    """Configure authentication

    This command will setup authentication profile(s). Note that this
    commnad can be used manually via terminal prompts, or via a JSON
    file containing one or more authentication info.

    Example --auth-file file, with single profile setup

    \b
    {
        "<PROFILE NAME>": {
            "active": <true|false>,
            "api_token": "<API TOKEN>",
            "jenkins_server_url": "<SERVER BASE URL>",
            "username": "<USERNAME ID>"
        }
    }
    """
    set_debug_log_level(debug)
    cli_auth.configure(auth_file)


@auth.command(short_help='\tGenerate authentication API token')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--token-name', type=str, required=False, help='Name of the generated API token')
@click.option('--server-base-url', type=str, required=False, help='Server base URL address')
@click.option('--username', type=str, required=False, help='Account username')
@click.option('--password', type=str, required=False, help='Account password')
def token(debug, profile, token_name, server_base_url, username, password):
    """Generate authentication API token

    NOTE: To generate an API token and automatically add it to the
    an existing authentication profile, use the --profile option.
    """
    set_debug_log_level(debug)
    cli_auth.token(profile, token_name, server_base_url, username, password)


@auth.command(short_help='\tShow the local credentials profiles')
@cli_decorators.debug
@cli_decorators.format_output
def show(debug, pretty, yaml, xml, toml):
    """Show the local credentials profiles"""
    set_debug_log_level(debug)
    cli_auth.show(pretty, yaml, xml, toml)


@auth.command(short_help='\tCheck if credentials can authenticate')
@cli_decorators.debug
@cli_decorators.profile
def verify(debug, profile):
    """Check if credentials can authenticate"""
    set_debug_log_level(debug)
    cli_auth.verify(profile)


@auth.command(short_help='\tWipe all credentials for this device')
@cli_decorators.debug
def wipe(debug):
    """Wipe all credentials for this device"""
    set_debug_log_level(debug)
    click.secho('TODO :-/', fg='yellow')


@auth.command(short_help='\tShow current user information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
def user(debug, pretty, yaml, xml, toml, profile):
    """Show current user information"""
    set_debug_log_level(debug)
    cli_auth.user(pretty, yaml, xml, toml, profile)



"""Job class definition"""

import json
import logging
import re
from datetime import timedelta
from time import perf_counter
from typing import Dict, Tuple, Union
from urllib.parse import urlencode

import jenkins
import xmltodict

from yojenkins.monitor import JobMonitor
from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out, failures_out
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses
from yojenkins.yo_jenkins.jenkins_item_config import JenkinsItemConfig

# Getting the logger reference
logger = logging.getLogger()


class Job():
    """TODO Job"""

    def __init__(self, rest, Folder, JenkinsSDK, auth, Build) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest
        self.folder = Folder
        self.jenkins_sdk = JenkinsSDK
        self.auth = auth
        self.build = Build
        self.JM = JobMonitor(rest, auth, self, Build)

        # Recursive search results
        self.search_results = []
        self.search_items_count = 0

    def _recursive_search(self, search_pattern: str, search_list: list, level: int, fullname: bool = True) -> None:
        """Recursive search method for jobs

        Details: Matched pattern findings are storred in the object: `self.search_results`

        Args:
            search_pattern : REGEX pattern to match for each item
            search_list    : List of items
            level          : Current recursion level
            fullname       : Search the entire path of the item, not just the item name

        Returns:
            None
        """
        # Current directory level
        level += 1

        # Loop through all sub-folders
        for list_item in search_list:

            # Check if it is not a job
            if list_item['_class'] in JenkinsItemClasses.JOB.value['class_type']:

                # Get fullname, else get name
                if fullname:
                    dict_key = "fullname" if "fullname" in list_item else "name"
                else:
                    dict_key = 'name'

                # Match the regex pattern
                try:
                    if re.search(search_pattern, list_item[dict_key], re.IGNORECASE):
                        self.search_results.append(list_item)
                except re.error as error:
                    logger.debug(
                        f'Error while applying REGEX pattern "{search_pattern}" to "{list_item[dict_key]}". Exception: {error}'
                    )
                    break

            # Count items searched for record
            self.search_items_count += 1

            # Check if it is a folder, if it is, keep looking
            if 'jobs' not in list_item:
                continue

            # Keep searching all sub-items for this item. Call itself for some recursion fun
            self._recursive_search(search_pattern, list_item['jobs'], level, fullname)

    def search(self,
               search_pattern: str,
               folder_name: str = '',
               folder_url: str = '',
               folder_depth: int = 4,
               fullname: bool = True) -> Tuple[list, list]:
        """TODO Docstring

        Args:
            arg 1 : DESCRIPTION
            arg 2 : DESCRIPTION

        Returns:
            TODO
        """
        # Finding the job by REGEX pattern
        # NOTE:
        #   - Criteria of jobs is that jobs do not have any sub-folders, only views and jobs

        # Start a timer to time the search
        start_time = perf_counter()

        logger.debug(f'Job search pattern: {search_pattern}')

        # Get all the jobs
        if folder_name or folder_url:
            # Only recursively search the specified folder name
            logger.debug(f'Searching jobs in sub-folder "{folder_name if folder_name else folder_url}"')
            logger.debug('Folder depth does not apply. Only looking in this specific folder for job')
            items = self.folder.item_list(folder_name=folder_name, folder_url=folder_url)[0]
        else:
            # Search entire Jenkins
            logger.debug(f'Searching jobs in ALL Jenkins. Folder depth: "{folder_depth}"')
            try:
                items = self.jenkins_sdk.get_all_jobs(folder_depth=folder_depth)
            except jenkins.JenkinsException as error:
                error_no_html = error.args[0].split("\n")[0]
                fail_out(f'Error while getting all items. Exception: {error_no_html}')
            except Exception as error:
                fail_out(error)

        # Search for any matching folders ("jobs")
        self.search_results = []
        self.search_items_count = 0
        self._recursive_search(search_pattern=search_pattern, search_list=items, level=0, fullname=fullname)

        # Remove duplicates from list
        logger.debug('Removing duplicates if needed ...')
        self.search_results = [i for n, i in enumerate(self.search_results) if i not in self.search_results[n + 1:]]

        # Getting only the URLs of the stages
        job_search_results_list = [result['url'] for result in self.search_results]

        # Output search stats
        logger.debug(
            f'Searched jobs: {self.search_items_count}. Search time: {perf_counter() - start_time:.3f} seconds')

        return self.search_results, job_search_results_list

    def info(self, job_name: str = '', job_url: str = '') -> Dict:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_name and not job_url:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Job url passed: {job_url}')
        job_info, _, success = self.rest.request(f'{job_url.strip("/")}/api/json', 'get', is_endpoint=False)
        if not success:
            fail_out(f'Failed to find job info: {job_url}')

        # Check if found item type/class
        if job_info['_class'] not in JenkinsItemClasses.JOB.value['class_type']:
            fail_out(f'Job found, but failed to match type/class. The found item is "{job_info["_class"]}"')

        if 'url' in job_info:
            job_info['fullName'] = utility.url_to_name(job_info['url'])
            job_info['jobUrl'] = utility.build_url_to_other_url(job_info['url'], target_url='job')
            job_info['jobFullName'] = utility.url_to_name(job_info['jobUrl'])
            job_info['folderUrl'] = utility.build_url_to_other_url(job_info['url'], target_url='folder')
            job_info['folderFullName'] = utility.url_to_name(job_info['folderUrl'])
            job_info['serverURL'] = utility.item_url_to_server_url(job_info['url'])
            job_info['serverDomain'] = utility.item_url_to_server_url(job_info['url'], False)

            job_info[
                'folderFullName'] = 'Base Folder' if not job_info['folderFullName'] else job_info['folderFullName']

        return job_info

    def build_list(self, job_name: str = '', job_url: str = '') -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Get the job information
        job_info = self.info(job_name=job_name, job_url=job_url)

        # Get all the past builds
        build_list, build_url_list = utility.item_subitem_list(
            item_info=job_info,
            get_key_info='url',
            item_type=JenkinsItemClasses.BUILD.value['item_type'],
            item_class_list=JenkinsItemClasses.BUILD.value['class_type'])

        return build_list, build_url_list

    def build_next_number(self, job_name: str = '', job_url: str = '') -> Union[int, None]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Get the job information
        job_info = self.info(job_name=job_name, job_url=job_url)

        if not job_info.get('nextBuildNumber'):
            fail_out('Failed to get next build number from job. "builds" key missing in job information')

        return job_info['nextBuildNumber']

    def build_last_number(self, job_name: str = '', job_url: str = '', job_info: dict = {}) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Return a url of the build instead just the number

        # Get the job information
        if not job_info:
            # If the job info is not passed, request it from server
            job_info = self.info(job_name=job_name, job_url=job_url)

        if not job_info.get('lastBuild'):
            return 0

        if 'lastBuild' not in job_info:
            fail_out('Failed to get last build number from job. "lastBuild" key missing in job information')
        if 'number' not in job_info['lastBuild']:
            fail_out('Failed to get last build number from job. "lastBuild.number" key missing in job information')

        return job_info['lastBuild']['number']

    def build_set_next_number(self, build_number: int, job_name: str = '', job_url: str = '') -> Union[int, None]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')
        if job_url and not job_name:
            job_name = utility.url_to_name(url=job_url)
        # Format name
        job_name = utility.format_name(name=job_name)

        logger.debug(f'Setting next build number for job "{job_name}" to {build_number} ...')

        try:
            # TODO: Use requests instead of jenkins-python
            self.jenkins_sdk.set_next_build_number(job_name, build_number)
        except jenkins.JenkinsException as error:
            error_no_html = error.args[0].split("\n")[0]
            fail_out(
                f'Failed to set next build number for job "{job_name}" to {build_number}. Exception: {error_no_html}')

        return build_number

    def build_number_exist(self,
                           build_number: int,
                           job_info: dict,
                           job_name: str = '',
                           job_url: str = '') -> Union[bool, None]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_info:
            # Getting job information
            job_info = self.info(job_name=job_name, job_url=job_url)

        if 'builds' not in job_info:
            fail_out('Failed to get build list from job. "builds" key missing in job information')

        # Iterate through all listed builds
        for build in job_info['builds']:
            if build_number == build['number']:
                return True

        return False

    def build_trigger(self, job_name: str = '', job_url: str = '', paramters: Dict = {}, token: str = '') -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # NOTE: The jenkins-python module build_job() does not work. Using requests instead

        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        logger.debug(f'Job reference passed: {job_name if job_name else job_url}')

        # Need both the job name and the job URL
        if job_url and not job_name:
            job_name = utility.url_to_name(url=job_url)
        elif job_name and not job_url:
            # TODO: Use requests instead of Jenkins SDK
            job_url = self.jenkins_sdk.build_job_url(name=job_name).strip('build')
        job_url = job_url.strip('/')

        next_build_number = self.build_next_number(job_name=job_name, job_url=job_url)
        logger.debug(f'Triggering job "{job_url}", build {next_build_number} ...')

        if paramters:
            # Use the paramters passed
            logger.debug(f'Triggering with job paramters: {paramters}')
            post_url = f'{job_url}/buildWithParameters?{urlencode(paramters)}'
        else:
            # No paramters passed
            post_url = f'{job_url}/build'

        logger.debug(f'POST url: {post_url}')

        # Posting to Jenkins
        return_headers = self.rest.request(post_url, 'post', is_endpoint=False)[1]

        # Parse the queue location of the build
        if return_headers:
            build_queue_url = return_headers['Location']
            if build_queue_url.endswith('/'):
                queue_location = build_queue_url[:-1]
            parts = queue_location.split('/')
            build_queue_number = int(parts[-1])
            logger.debug(f'Build queue URL: {queue_location}')
            logger.debug(f'Build queue ID: {build_queue_number}')
        else:
            fail_out('Failed to trigger build. Failed to send request')

        return build_queue_number

    def wipeout_workspace(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO
        pass

    def queue_info(self, build_queue_number: int = 0, build_queue_url: str = '') -> Dict:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Getting information for build queue "{build_queue_number}" ...')

        # Make the request URL
        if build_queue_number:
            endpoint = f'queue/item/{build_queue_number}/api/json'
        elif build_queue_url:
            endpoint = f'{build_queue_url}api/json'
        else:
            fail_out('No build queue number or build queue url passed')
        queue_info = self.rest.request(endpoint, 'get', is_endpoint=True)[0]

        # Adding additional parameters
        if queue_info:
            queue_info['isQueuedItem'] = True
            queue_info['fullUrl'] = self.rest.get_server_url().strip('/') + '/' + queue_info['url']
            queue_info['jobUrl'] = queue_info['task']['url']
            queue_info['jobFullName'] = utility.url_to_name(queue_info['jobUrl'])
            queue_info['folderUrl'] = utility.build_url_to_other_url(queue_info['fullUrl'], target_url='folder')
            queue_info['folderFullName'] = utility.url_to_name(queue_info['folderUrl'])
            queue_info['serverURL'] = utility.item_url_to_server_url(queue_info['url'])
            queue_info['serverDomain'] = utility.item_url_to_server_url(queue_info['url'], False)

        return queue_info

    def in_queue_check(self, job_name: str = '', job_url: str = '') -> Tuple[dict, int]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        # Requesting all queue and searching queue (NOTE: Could use Server object)
        queue_all = self.rest.request('queue/api/json', 'get')[0]
        logger.debug(f"Number of queued items: {len(queue_all['items'])}")
        queue_matches = utility.queue_find(queue_all, job_name=job_name, job_url=job_url)
        if not queue_matches:
            return {}, 0
        queue_info = queue_matches[0]

        # Adding additional parameters
        if queue_info:
            queue_info['inQueueSinceFormatted'] = str(timedelta(seconds=queue_info['inQueueSince'] / 1000.0))[:-3]
            queue_info['fullUrl'] = self.rest.get_server_url() + '/' + queue_info['url']
            queue_info['jobUrl'] = queue_info['task']['url']
            queue_info['jobFullName'] = utility.url_to_name(queue_info['jobUrl'])
            queue_info['folderUrl'] = utility.build_url_to_other_url(queue_info['fullUrl'], target_url='folder')
            queue_info['folderFullName'] = utility.url_to_name(queue_info['folderUrl'])
            queue_info['serverURL'] = utility.item_url_to_server_url(queue_info['fullUrl'])
            queue_info['serverDomain'] = utility.item_url_to_server_url(queue_info['fullUrl'], False)

        return queue_info, queue_info['id']

    def queue_abort(self, build_queue_number: int) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Aborting build queue "{build_queue_number}" ...')

        if not build_queue_number:
            fail_out('No build queue number passed')

        # Make the request URL
        endpoint = f'queue/cancelItem?id={build_queue_number}'
        return_content = self.rest.request(endpoint, 'post', is_endpoint=True)[0]

        if not return_content:
            messages = [
                'Failed to abort build queue. Specified build queue number may be wrong or build may have already started',
                'The following jobs are currently in queue:'
            ]
            queue_list = self.in_queue_check()
            for i, queue_item in enumerate(queue_list):
                messages.append(f'  {i+1}. Queue ID: {queue_item["id"]} - Job URL: {queue_item["task"]["url"]}')
            failures_out(messages)

        return True

    def browser_open(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Opening in web browser: "{job_url}" ...')
        if not utility.browser_open(url=job_url):
            fail_out('Failed to open job in web browser')
        logger.debug('Successfully oped job in web browser')

        return True

    def config(self,
               filepath: str = '',
               job_name: str = '',
               job_url: str = '',
               opt_json: bool = False,
               opt_yaml: bool = False,
               opt_toml: bool = False) -> str:
        """Get the folder XML configuration (config.xml)

        Args:
            filepath    : If defined, store fetched data in this file
            job_name    : Job name to get configurations
            folder_url  : Job URL to get configurations

        Returns:
            Folder config.xml contents
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Fetching XML configurations for job: "{job_url}" ...')
        return_content, _, success = self.rest.request(f'{job_url.strip("/")}/config.xml',
                                                       'get',
                                                       json_content=False,
                                                       is_endpoint=False)
        if not success:
            fail_out('Failed to get job configuration')
        logger.debug('Successfully fetched XML configurations')

        if filepath:
            write_success = utility.write_xml_to_file(return_content, filepath, opt_json, opt_yaml, opt_toml)
            if not write_success:
                fail_out('Failed to write configuration file')

        return return_content

    def disable(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Disabling job: "{job_url}" ...')
        success = self.rest.request(f'{job_url.strip("/")}/disable', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to disable job')
        logger.debug('Successfully disabled job')

        return success

    def enable(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Enabling job: "{job_url}" ...')
        success = self.rest.request(f'{job_url.strip("/")}/enable', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to enable job')
        logger.debug('Successfully enabled job')

        return success

    def rename(self, new_name: str, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        if not new_name:
            fail_out('The new job name is a blank')
        if utility.has_special_char(new_name):
            fail_out('The new job name contains special characters')

        logger.debug(f'Renaming job: "{job_url}" ...')
        success = self.rest.request(f'{job_url.strip("/")}/doRename?newName={new_name}', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to rename job')
        logger.debug('Successfully renamed job')

        return success

    def delete(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Deleting job: "{job_url}" ...')
        success = self.rest.request(f'{job_url.strip("/")}/doDelete', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to delete job')
        logger.debug('Successfully deleted job')

        return success

    def wipe_workspace(self, job_name: str = '', job_url: str = '') -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        logger.debug(f'Wiping workspace for job: "{job_url}" ...')
        success = self.rest.request(f'{job_url.strip("/")}/doWipeOutWorkspace', 'post', is_endpoint=False)[2]
        if not success:
            fail_out('Failed to wipe workspace')
        logger.debug('Successfully wiped workspace')

        return success

    def monitor(self, job_name: str = '', job_url: str = '', sound: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not job_name and not job_url:
            fail_out('No job name or job URL provided')

        if job_url:
            job_url = job_url.strip('/')
        else:
            job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

        if not self.rest.request(f'{job_url.strip("/")}/api/json', 'head', is_endpoint=False)[2]:
            fail_out(f'Failed to find job. The job may not exist: {job_url}')

        logger.debug(f'Starting monitor for: "{job_url}" ...')
        success = self.JM.monitor_start(job_url=job_url, sound=sound)
        if not success:
            fail_out('Failed to start job monitor')
        logger.debug('Successfully started job monitor')

        return success

    def create(self,
               name: str,
               folder_name: str = '',
               folder_url: str = '',
               config_file: str = 'config.xml',
               config_is_json: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if not folder_name and not folder_url:
            fail_out('No folder name or folder url received')

        if folder_url:
            folder_url = folder_url.strip('/')
        else:
            folder_url = utility.name_to_url(self.rest.get_server_url(), folder_name)

        if not name:
            fail_out('Provided item name is a blank')
        if utility.has_special_char(name):
            fail_out('Provided item name contains special characters')

        # Check if job already exists
        if utility.item_exists_in_folder(name, folder_url, "job", self.rest):
            fail_out(f'Job "{name}" already exists in folder "{folder_url}"')

        if config_file:
            # Use job config from file
            logger.debug(f'Opening and reading file: {config_file} ...')
            try:
                open_file = open(config_file, 'rb')
                job_config = open_file.read()
            except (OSError, IOError, PermissionError) as error:
                fail_out(f'Failed to open and read file. Exception: {error}')

            if config_is_json:
                logger.debug('Converting the specified JSON file to XML format ...')
                try:
                    job_config = xmltodict.unparse(json.loads(job_config))
                except ValueError as error:
                    fail_out(f'Failed to convert the specified JSON file to XML format. Exception: {error}')
        else:
            # Use blank job config template
            job_config = JenkinsItemConfig.JOB.value['blank']

        logger.debug(f'Creating job "{name}" within folder "{folder_url}" "...')
        endpoint = f'createItem?name={name}'
        headers = {'Content-Type': 'application/xml; charset=utf-8'}
        _, _, success = self.rest.request(f'{folder_url.strip("/")}/{endpoint}',
                                          'post',
                                          data=job_config,
                                          headers=headers,
                                          is_endpoint=False)
        if not success:
            fail_out(f'Failed to create job "{name}"')
        logger.debug(f'Successfully created job "{name}"')

        try:
            if 'open_file' in locals():
                open_file.close()
        except (OSError, IOError):
            pass

        return success

    def parameters(self, job_name: str = '', job_url: str = '') -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Get the job information
        job_info = self.info(job_name=job_name, job_url=job_url)

        logger.debug(f'Getting build parameters for job: "{job_name}" ...')

        # Check if item has any parameter actions
        parameter_actions = utility.get_item_action(job_info, 'hudson.model.ParametersDefinitionProperty')
        if not parameter_actions:
            fail_out('This job does not have any build parameters')

        # Get the parameter definitions
        parameters = parameter_actions[0]['parameterDefinitions']

        # List of parameter items
        parameters_list = []
        for parameter in parameters:
            # Default value of parameter
            default_value = parameter['defaultParameterValue']['value']
            default_value_str = ''
            if not default_value:
                if parameter['defaultParameterValue']['_class'] == 'hudson.model.BooleanParameterValue':
                    default_value_str = f" (default: False)"
            else:
                default_value_str = f" (default: {default_value})"

            # Description of parameter
            parameter_description = parameter['description'] if parameter['description'] else 'N/A'

            item = f'{parameter["type"]} - {parameter["name"]} - {parameter_description}{default_value_str}'
            parameters_list.append(item)

        return parameters, parameters_list



"""Utility/Tools Menu CLI Entrypoints"""

import json
import logging
import os
import platform
import sys
from datetime import datetime
from inspect import getfullargspec
from pathlib import Path
from shlex import quote
from typing import Callable

import click
import toml
import yaml
from json2xml import json2xml
from json2xml.utils import readfromstring
from urllib3.util import parse_url

from yojenkins import __version__
from yojenkins.yo_jenkins.auth import Auth
from yojenkins.yo_jenkins.rest import Rest
from yojenkins.yo_jenkins.yojenkins import YoJenkins

from yojenkins.utility.utility import iter_data_empty_item_stripper, load_contents_from_local_file, am_i_inside_docker, am_i_bundled, print2  # isort:skip

# Getting the logger reference
logger = logging.getLogger()

CONFIG_DIR_NAME = '.yojenkins'
HISTORY_FILE_NAME = 'history'
COMMAND_HISTORY_FORMAT = 'json'
DEFAULT_PROFILE_NAME = 'default'
MAX_PROFILE_HISTORY_LENGTH = 1000

CLI_CMD_PATH = sys.argv[0]
CLI_CMD_ARGS = ' '.join([quote(arg) for arg in sys.argv[1:]])


def set_debug_log_level(debug_flag: bool) -> None:
    """Setting the log DEBUG level

    Args:
        debug_flag : Boolean flag, True to set to DEBUG, else INFO

    Returns:
        None
    """
    if debug_flag:
        logging_level = logging.DEBUG
        click.secho('\n[ LOGGING LEVEL ] : DEBUG\n', fg='bright_yellow', bold=True)
        line = '[  TIME  ] [ MS ] [FILENAME              :  LN] MESSAGE'
        click.secho(line, fg='bright_yellow', bold=True)
        line = '---------------------------------------------------------------------------'
        click.secho(line, fg='bright_yellow', bold=True)
    else:
        logging_level = logging.INFO
    logger.setLevel(logging_level)
    for handler in logger.handlers:
        handler.setLevel(logging_level)

    # Program version
    logger.debug(f'Tool version: {__version__}')

    # Show system information
    platform_information()


def platform_information() -> None:
    """Display (Console out) current system and program information
    """
    python_rev = f'(REV:{platform.python_revision()})' if platform.python_revision() else ''

    # System information
    logger.debug('System information:')
    logger.debug(f'    - System:    {platform.system()}')
    logger.debug(f'    - Release:   {platform.uname().release}')
    logger.debug(f'    - Version:   {platform.version()}')
    logger.debug(f'    - Machine:   {platform.uname().machine}')
    logger.debug(f'    - Processor: {platform.uname().processor}')
    logger.debug(f'    - Python:    {platform.python_version()} {python_rev}')
    logger.debug(f'    - In Docker: {am_i_inside_docker()}')
    logger.debug(f'    - Bundled:   {am_i_bundled()}')


def config_yo_jenkins(profile: str) -> YoJenkins:
    """TODO Docstring

    Args:
        TODO

    Returns:
        TODO
    """
    auth = Auth(Rest())

    # Get the credential profile
    if not auth.get_credentials(profile):
        click.secho('Failed to find any credentials', fg='bright_red', bold=True)
        sys.exit(1)

    # Create authentication
    if not auth.create_auth():
        click.secho('Failed authentication', fg='bright_red', bold=True)
        sys.exit(1)

    return YoJenkins(auth)


def standard_out(data: dict,
                 opt_pretty: bool = False,
                 opt_yaml: bool = False,
                 opt_xml: bool = False,
                 opt_toml: bool = False) -> None:
    """TODO Docstring

    Args:
        TODO

    Returns:
        TODO
    """
    # Strip away any empty items in the iterable data
    logger.debug('Removing all empty items in iterable data ...')
    data = iter_data_empty_item_stripper(data)

    if opt_pretty:
        logger.debug('"PRETTY" (human readable) output was enabled')

    if opt_xml:
        logger.debug('Outputting XML format ...')
        if isinstance(data, dict) or isinstance(data, list):
            data = readfromstring(json.dumps(data))
            data_xml = json2xml.Json2xml(data, pretty=opt_pretty, wrapper=None, attr_type=False).to_xml()
            if opt_pretty:
                print2(data_xml)
            else:
                print2(data_xml.decode())
        else:
            # When configs are fetched in XML format
            print2(data)
        return

    if opt_yaml:
        # YAML format
        logger.debug('Outputting YAML format ...')
        print2(yaml.safe_dump(data, default_flow_style=False, indent=2))
    elif opt_toml:
        # TOML format
        data = {'item': data} if isinstance(data, list) else data
        logger.debug('Outputting TOML format ...')
        print2(toml.dumps(data))
    else:
        # JSON format
        logger.debug('Outputting JSON format ...')
        if opt_pretty:
            print2(json.dumps(data, indent=4, sort_keys=True))
        else:
            print2(json.dumps(data))


def is_full_url(url: str) -> bool:
    """TODO Docstring

    Args:
        TODO

    Returns:
        TODO
    """

    # TODO: Remove this function from this file
    #       Do url check within the class, not within the cli to not keep repeating it
    #       In classes use yojenkins.utility.utility.is_full_url()

    parsed_url = parse_url(url)
    if all([parsed_url.scheme, parsed_url.netloc, parsed_url.path]):
        is_valid_url = True
    else:
        is_valid_url = False
    logger.debug(f'Is valid URL format: {is_valid_url} - {url}')
    logger.debug(f'    - scheme:  {parsed_url.scheme} - {"OK" if parsed_url.scheme else "MISSING"}')
    logger.debug(f'    - netloc:  {parsed_url.netloc} - {"OK" if parsed_url.netloc else "MISSING"}')
    logger.debug(f'    - path:    {parsed_url.path} - {"OK" if parsed_url.path else "MISSING"}')

    return is_valid_url


def server_target_check(target: str) -> bool:
    """TODO"""
    pass


def log_to_history(decorated_function) -> Callable:
    """This function decorates a function that is a cli command.
    The function will log the CLI command and its info to the command history.

    Details: Add this function like "@log_to_history" to a function

    Args:
        decorated_function : Function that is decorated

    Returns:
        None
    """
    # Get the profile argument index
    argspec = getfullargspec(decorated_function)
    try:
        arg_index = argspec.args.index('profile')
    except ValueError:
        arg_index = -1

    def wrapper(*args, **kwargs) -> None:
        # Get the profile name for the command
        if arg_index >= 0:
            profile_name = args[arg_index]
        else:
            # NOTE: If no profile is used by the decorated function, use the default profile name
            profile_name = DEFAULT_PROFILE_NAME
        if profile_name is None:
            # NOTE: If function has profile argument, but none was passed, use the default profile name
            profile_name = DEFAULT_PROFILE_NAME

        # Check if history file exists
        history_file_path = os.path.join(os.path.join(Path.home(), CONFIG_DIR_NAME), HISTORY_FILE_NAME)
        if not os.path.isfile(history_file_path):
            logger.debug(f'Failed to find command history file: "{history_file_path}"')
            logger.debug('Creating blank command history file ...')
            try:
                with open(history_file_path, 'w') as open_file:
                    json.dump({}, open_file)
            except Exception as error:
                logger.debug(f'Failed to create new command history file: {error}')

        # Load the history file content
        file_contents = load_contents_from_local_file(COMMAND_HISTORY_FORMAT, history_file_path)
        if not file_contents:
            logger.debug("Command history file is blank")

        # If profile history length is too long, remove the oldest history item
        if profile_name in file_contents:
            if len(file_contents[profile_name]) > MAX_PROFILE_HISTORY_LENGTH:
                file_contents[profile_name].pop(0)

        # Add the command to the history file
        logger.debug(f'Logging command to command history file: "{history_file_path}" ...')
        if profile_name not in file_contents:
            file_contents[profile_name] = []

        command_info = {
            'tool_path': CLI_CMD_PATH,
            'arguments': CLI_CMD_ARGS,
            'timestamp': datetime.now().timestamp(),
            'datetime': datetime.now().strftime("%A, %B %d, %Y %I:%M:%S"),
            'tool_version': __version__
        }
        file_contents[profile_name].append(command_info)

        # Add to file, overwritting entire file
        try:
            with open(history_file_path, 'w') as outfile:
                json.dump(file_contents, outfile, indent=4)
        except Exception as error:
            logger.debug(f'Failed to write command history file: {error}')

        return decorated_function(*args, **kwargs)

    return wrapper



"""Project dynamic management"""

import logging
import subprocess
import sys

# Getting the logger reference
logger = logging.getLogger()

# NOTE: Using pip within code with "import pip" is not recommended:
#       https://pip.pypa.io/en/latest/user_guide/#using-pip-from-your-program


class Package():
    """Class managing updating of this software"""

    @staticmethod
    def install(package_name: str = 'yojenkins', upgrade: bool = True, user: bool = True, proxy: str = "") -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Installing "{package_name}" ...')
        logger.debug(f'    - Upgrade:      {upgrade}')
        logger.debug(f'    - User context: {user}')
        logger.debug(f'    - Proxy:        {proxy if proxy else "None"}')

        upgrade_option = "--upgrade" if upgrade else ""
        user_option = "--user" if user else ""
        proxy_option = proxy if proxy else ""

        arg_list = [sys.executable, "-m", "pip", "install", upgrade_option, user_option, proxy_option, package_name]

        # Remove blanks
        arg_list = [arg_item for arg_item in arg_list if arg_item]

        try:
            subprocess.check_call(arg_list)
        except Exception as error:
            logger.debug(f'Failed to install/upgrade {package_name}. Exception: {error}')
            return False

        return True

    @staticmethod
    def uninstall(package_name: str = 'yojenkins', auto_yes: bool = True) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Uninstalling "{package_name}" ...')
        logger.debug(f'    - Auto yes: {auto_yes}')

        auto_yes_option = "--yes" if auto_yes else ""

        arg_list = [sys.executable, "-m", "pip", "uninstall", auto_yes_option, package_name]

        # Remove blanks
        arg_list = [arg_item for arg_item in arg_list if arg_item]

        try:
            subprocess.check_call(arg_list)
        except Exception as error:
            logger.debug(f'Failed to uninstall {package_name}. Exception: {error}')
            return False
        return True



"""Jenkins item type/class definition"""

from enum import Enum


class JenkinsItemClasses(Enum):
    """Enum of Jenkins classes ordered in categories for checking and comparison

    item_type  - How it is stored in the return JSON object from server. The name of the sub-section
    prefix     - The prefix in the URL (ie. for view -> http://localhost:8080/view/My_View/)
    class_type - Any class type Jenkins categorizes this item

    Usage Examples:
        - `JenkinsItemClasses.FOLDER.value['class_type']`
        - `JenkinsItemClasses.VIEW.value['item_type']`
    """
    FOLDER = {
        "item_type":
            "jobs",
        "prefix":
            "job",
        "class_type": [
            'com.cloudbees.hudson.plugins.folder.Folder',
            'org.jenkinsci.plugins.workflow.multibranch.WorkflowMultiBranchProject',
            'jenkins.branch.OrganizationFolder'
        ]
    }
    VIEW = {
        "item_type":
            "views",
        "prefix":
            "view",
        "class_type": [
            'hudson.model.AllView', 'hudson.model.ListView', 'jenkins.branch.MultiBranchProjectViewHolder$ViewImpl'
        ]
    }
    JOB = {
        "item_type": "jobs",
        "prefix": "job",
        "class_type": ['org.jenkinsci.plugins.workflow.job.WorkflowJob', 'hudson.model.FreeStyleProject']
    }
    BUILD = {
        "item_type": "builds",
        "prefix": "",
        "class_type": ['org.jenkinsci.plugins.workflow.job.WorkflowRun', 'hudson.model.FreeStyleBuild']
    }
    QUEUE = {
        "item_type": "queue",
        "prefix": "",
        "class_type": ['hudson.model.Queue$BuildableItem', 'hudson.model.Queue$BlockedItem']
    }
    NODE = {
        "item_type": "computer",
        "prefix": "",
        "class_type": ['hudson.slaves.SlaveComputer', 'hudson.slaves.DumbSlave', 'hudson.model.Hudson$MasterComputer']
    }



"""Job monitor"""

import curses
import logging
import sys
import threading
from datetime import datetime
from time import perf_counter, sleep, time

from yojenkins.monitor.monitor import Monitor
from yojenkins.yo_jenkins.status import BuildStatus

from . import monitor_utility as mu

# Getting the logger reference
logger = logging.getLogger()


class JobMonitor(Monitor):
    """This class defines the JobMonitor class and its function.

    The JobMonitor class enables active job monitoring
    """

    def __init__(self, rest, auth, Job, Build) -> None:
        """Object constructor method, called at object creation

        Args:
            Build: Build object

        Returns:
            None
        """
        # Get attributes from super (parent) class
        super().__init__()

        self.rest = rest
        self.auth = auth
        self.job = Job
        self.build = Build

        # Job Info Thread
        self.job_info_data = {}
        self.job_info_thread_interval = 0.0
        self._job_info_thread_lock = threading.Lock()

        # Builds Thread
        self.builds_data = []
        self.builds_data_number_of_builds = 10
        self.builds_data_thread_interval = 0.0
        self._build_data_thread_lock = threading.Lock()

        # Building a job flag
        self.job_build = 0

        # Temporary message box on screen
        self.message_box_temp_duration = 1  # sec

    ###########################################################################
    #                         BUILD MONITOR
    ###########################################################################

    def __monitor_draw(self, scr, job_url: str, sound: bool = False) -> bool:
        """
        Draw a the BUILD MONITOR UI on the screen

        Args:
            scr       : Handle for curses terminal screen handle
            job_url : Direct URL to build
            sound     : Enable sound effects
        Returns:
            True if no error, else False
        """
        # Starting data collection threads
        self.server_status_thread_on()
        self.__job_info_thread_on(job_url=job_url)
        self.__builds_data_thread_on()

        # Setting up basic stuff for curses and load keys
        self.basic_screen_setup(halfdelay=True)
        ui_keys = mu.load_keys()

        # User key input (ASCII value)
        keystroke = 0

        # Debug
        loop_total_time = 0

        # Main Loop
        while True:
            start_time = perf_counter()

            # Clearing the screen at each loop iteration before constructing the frame
            scr.clear()

            ########################################################################################

            # Check user keyboard input
            if keystroke in ui_keys['QUIT']:
                self.quit += 1
            elif keystroke in ui_keys['BUILD']:
                self.job_build += 1
            elif keystroke in ui_keys['RESUME']:
                self.quit = 0
                self.job_build = 0
            elif keystroke in ui_keys['PAUSE']:
                self.paused = not self.paused
            elif keystroke in ui_keys['HELP']:
                self.help = not self.help
            elif keystroke in ui_keys['OPEN']:
                if job_url:
                    self.job.browser_open(job_url=job_url)

            ########################################################################################

            # Check terminal size
            term_height, term_width = scr.getmaxyx()
            self.check_terminal_size(scr)

            # Paint background
            mu.paint_background(scr, self.color['normal'])

            ########################################################################################

            # TOP HEADER
            y_row = 1

            if logger.level < 20:
                pass
            else:
                mu.draw_text(scr,
                             'JOB MONITOR',
                             y_row,
                             center_x=True,
                             color=self.color['grey-light'],
                             decor=self.decor['bold'])

            y_row += 1

            # Draw header divider
            mu.draw_horizontal_seperator(scr, y_row, self.color['grey-dark'])
            y_row += 2

            ########################################################################################

            # INFO SECTION
            x_col = [3, 16]
            if self.job_info_data:
                # Get the job_url
                job_url = self.job_info_data['url']

                # INFO
                mu.draw_horizontal_header(scr, y_row, x_col[0], term_width - 5, '-', 'INFO',
                                          self.color['normal'] | self.decor['bold'])
                y_row += 1

                job_info_items = {
                    'Job': self.job_info_data['displayName'],
                    'Folder': self.job_info_data['folderFullName'],
                    'Server': self.job_info_data['serverURL'],
                }
                for i, (key, value) in enumerate(job_info_items.items()):
                    mu.draw_text(scr, f'{key}:', y_row, x_col[0], decor=self.decor['bold'])
                    mu.draw_text(scr, mu.truncate_text(f'{value}', term_width - 5 - 12), y_row, x_col[1])
                    y_row += 1
                y_row += 1
            else:
                y_row += 3
                mu.draw_text(scr,
                             'NO DATA',
                             y_row,
                             center_x=True,
                             color=self.color['normal'],
                             decor=self.decor['bold'])
                y_row += 2
                mu.draw_text(scr,
                             'ಠ_ಠ  ¯\_(⊙︿⊙)_/¯',
                             y_row,
                             center_x=True,
                             color=self.color['normal'],
                             decor=self.decor['bold'])
            y_row += 1

            ########################################################################################

            # BUILDS SECTION
            if self.job_info_data and self.builds_data:
                x_col = [3, 12, 32, 52]

                # Header
                mu.draw_horizontal_header(scr, y_row, x_col[0], term_width - 5, '-', 'BUILDS',
                                          self.color['normal'] | self.decor['bold'])
                y_row += 1

                # Loop through all listed builds
                for i, build in enumerate(self.builds_data):
                    if not build: break

                    # Build name
                    line = build["displayName"] if "displayName" in build else build["number"]
                    mu.draw_text(scr, f'{build["displayName"]}', y_row, x_col[0])

                    # Datetime
                    line = datetime.fromtimestamp(build['timestamp'] / 1000.0).strftime("%m/%d - %H:%M")
                    mu.draw_text(scr, line, y_row, x_col[1])

                    # Build Run duration
                    if build['durationFormatted'] != None:
                        line = build['durationFormatted']
                    else:
                        line = build['elapsedFormatted']
                    mu.draw_text(scr, line, y_row, x_col[2])

                    # Build Status
                    if 'resultText' in build and build['resultText'] != None:
                        line = build['resultText']
                    else:
                        line = BuildStatus.UNKNOWN.value
                    status_color = self.status_to_color(line)
                    mu.draw_text(scr, line, y_row, x_col[3], color=self.color[status_color])

                    # Return down
                    y_row += 1
            else:
                # Change the minimum window height limit (no stages section)
                self.height_limit = 17

            # Divider
            y_row = term_height - 4
            mu.draw_horizontal_seperator(scr, y_row, self.color['grey-dark'])

            ########################################################################################

            # SERVER STATUS
            y_row = term_height - 3
            if self.server_status_data:
                auth_status = False if 'auth' not in self.server_status_data else self.server_status_data["auth"]
                reach_status = False if 'reachable' not in self.server_status_data else self.server_status_data[
                    "reachable"]
                line = f'Server Status: Reachable: {reach_status}, Authenticated: {auth_status}'
            else:
                line = 'Server Status: NO DATA'
            mu.draw_text(scr, line, y_row, center_x=True, color=self.color['grey-dark'])

            ########################################################################################

            # User key input instructions
            y_row = term_height - 2
            mu.draw_text(scr, 'Press "H" for keyboard shortcuts', y_row, center_x=True, color=self.color['grey-dark'])

            ########################################################################################

            # Drawing the screen border
            border_color = 'grey-dark'
            mu.draw_screen_border(scr, self.color[border_color])

            ########################################################################################

            # Indicate server interaction
            if self.server_interaction:
                mu.draw_text(scr,
                             '(R)',
                             term_height - 2,
                             term_width - 5,
                             color=self.color['grey-dark'],
                             decor=self.decor['bold'])
            self.server_interaction = False

            ########################################################################################

            # Refresh the screen (scr.nourefresh + scr.doupdate)
            scr.refresh()

            ########################################################################################

            halfdelay_normal = False

            # Help message box
            if self.help:
                curses.halfdelay(255)
                message_lines = [
                    'B - Build the job', 'O - Open job in web browser', 'P - Pause Monitor', 'Q - Quit Monitor', ' ',
                    'H - Keyboard shortcuts'
                ]
                mu.draw_message_box(scr, message_lines, 'left')
            else:
                halfdelay_normal = True

            # Pause message box
            if self.paused:
                self.help = False
                curses.halfdelay(255)
                message_lines = ['Monitor paused', 'Requests stopped', 'To resume press "P"']
                mu.draw_message_box(scr, message_lines)
            else:
                halfdelay_normal = True

            # Build the monitored job
            if self.job_build:
                self.help = False
                curses.halfdelay(255)
                message_lines = [
                    'Are you sure you want to build this job?', 'To build press "B"', 'To return press "R"'
                ]
                mu.draw_message_box(scr, message_lines)
                if self.job_build > 1:  # Abort Message confirmed (pressed twice)
                    if job_url:
                        self.server_interaction = True
                        self.job.build_trigger(job_url=job_url)
                    else:
                        pass
                    self.job_build = 0
            else:
                halfdelay_normal = True

            # Quit message box
            if self.quit:
                self.help = False
                curses.halfdelay(255)
                message_lines = ['Are you sure you want to quit?', 'To quit press "Q"', 'To return press "R"']
                mu.draw_message_box(scr, message_lines)
                # Quit Message confirmed (pressed twice)
                if self.quit > 1:
                    self.all_threads_enabled = False
                    return True
            else:
                halfdelay_normal = True

            # Screen refresh/updating to normal
            if halfdelay_normal:
                curses.halfdelay(self.halfdelay_screen_refresh)

            # Straight exist program
            if self.exit:
                self.all_threads_enabled = False
                sys.exit(0)

            ########################################################################################

            loop_total_time = perf_counter() - start_time

            # Get User input
            keystroke = scr.getch()

    def monitor_start(self, job_url: str, sound: bool = False) -> bool:
        """
        Curses wrapper function for drawing main menu on screen

        Args:
            build_url: Server URL of the build
            sound: If True, monitor is started with sound option on

        Returns:
            True, if successful, else False
        """
        # Disable any console output logging
        mu.logging_console(enabled=False)

        return curses.wrapper(self.__monitor_draw, job_url, sound)

    ###########################################################################
    #                      DATA COLLECTION THREADS
    ###########################################################################

    #############################  JOB INFO  ##################################

    def __thread_job_info(self, job_url: str, monitor_interval: float) -> None:
        """
        Independent thread which polls job information

        Args:
            job_url: Server URL of the job
            monitor_interval: Seconds between polling job URL

        Returns:
            None
        """
        logger.debug(
            f'Thread starting - Job info - (ID: {threading.get_ident()} - Refresh Interval: {monitor_interval}s) ...')

        # Set the monitoring thread flag up
        self.all_threads_enabled = True
        self.job_info_thread_interval = monitor_interval

        # Loop until flags disable it
        while self.all_threads_enabled:
            if not self.paused:
                self.server_interaction = True
                with self._job_info_thread_lock:
                    self.job_info_data = self.job.info(job_url=job_url)

            # Wait some time before checking again
            start_time = time()
            while time() - start_time <= monitor_interval:
                sleep(0.100)
                if not self.all_threads_enabled:
                    break

        logger.debug(f'Thread stopped - Job info - (ID: {threading.get_ident()})')

    def __job_info_thread_on(self, job_url: str = '', monitor_interval: float = 5.0) -> bool:
        """
        Trigger to start independent thread that polls job information

        Args:
            job_url: Server URL of the job
            monitor_interval: Seconds between polling build URL

        Returns:
            True if successful, else False
        """
        logger.debug(f'Starting thread for job info for "{job_url}" ...')
        try:
            threading.Thread(target=self.__thread_job_info, args=(
                job_url,
                monitor_interval,
            ), daemon=False).start()
        except Exception as error:
            logger.error(
                f'Failed to start job info monitoring thread for {job_url}. Exception: {error}. Type: {type(error)}')

        return True

    ############################  BUILDS INFO  ################################

    def __thread_build_info(self, build_url: str, build_data_index: int) -> None:
        """
        Independent thread which fetches build information

        Args:
            build_url: Server URL of the build
            build_data_index: Build index within the job info return

        Returns:
            None
        """
        logger.debug(f'Thread starting - Build info (INDEX: {build_data_index}, ID: {threading.get_ident()}) ...')
        self.server_interaction = True
        self.builds_data[build_data_index] = self.build.info(build_url=build_url)
        logger.debug(f'Thread stopped - Build info (INDEX: {build_data_index}, ID: {threading.get_ident()})')

    def __thread_builds_data(self, monitor_interval: float) -> None:
        """
        Independent thread which polls the build data for each build listed in job

        Args:
            monitor_interval: Seconds between polling build URL

        Returns:
            None
        """
        logger.debug(
            f'Thread starting - Builds data - (ID: {threading.get_ident()} - Refresh Interval: {monitor_interval}s) ...'
        )

        # Set the monitoring thread flag up
        self.all_threads_enabled = True
        self.builds_data_thread_interval = monitor_interval

        # Pre-allocate
        self.builds_data = [None] * self.builds_data_number_of_builds

        # Wait for initial batch of job info data
        while not self.job_info_data:
            if not self.all_threads_enabled:
                break
            sleep(0.100)

        # Loop until flags disable it
        while self.all_threads_enabled:
            if not self.paused:
                if 'builds' in self.job_info_data:
                    # Get and store the data in each thread
                    threads = []
                    try:
                        for build_data_index, build in enumerate(
                                self.job_info_data['builds'][:self.builds_data_number_of_builds]):
                            # build_data_index += 1
                            thread = threading.Thread(target=self.__thread_build_info,
                                                      args=(
                                                          build['url'],
                                                          build_data_index,
                                                      ),
                                                      daemon=False)
                            thread.start()
                            threads.append(thread)
                    except Exception as error:
                        logger.debug(f'Failure occurred when starting build data threads. Exception: {error}')

                    try:
                        for thread in threads:
                            thread.join()
                    except Exception as error:
                        logger.debug(f'Failure occurred when ending build data threads. Exception: {error}')
                else:
                    logger.debug('No job info data. Waiting ...')

            # Wait some time before checking again
            start_time = time()
            while time() - start_time <= monitor_interval:
                sleep(0.100)
                if not self.all_threads_enabled:
                    break

        logger.debug(f'Thread stopped - Builds data - (ID: {threading.get_ident()})')

    def __builds_data_thread_on(self, monitor_interval: float = 7.0) -> bool:
        """
        Trigger to start independent thread that polls build information within job

        Args:
            monitor_interval: Seconds between polling build data

        Returns:
            True if successful, else False
        """
        logger.debug('Starting thread for job builds info ...')
        try:
            threading.Thread(target=self.__thread_builds_data, args=(monitor_interval, ), daemon=False).start()
        except Exception as error:
            logger.error(f'Failed to start job builds info monitoring thread. Exception: {error}. Type: {type(error)}')

        return True
